@@ -1,7 +1,7 @@
 from __future__ import annotations
 
 from enum import Enum
-from typing import Dict, List, Optional, Set
+from typing import Any, Dict, List, Optional, Union
 
 from pydantic import BaseModel
 
@@ -43,6 +43,10 @@ class FileColumnMeaningType(Enum):
     EVAL_SET_INDEX = "EVAL_SET_INDEX"
     ENTITY_SYSTEM_RECORD_ID = "ENTITY_SYSTEM_RECORD_ID"
     UNNEST_KEY = "UNNEST_KEY"
+    IP_BINARY = "IP_BINARY"
+    IP_PREFIX = "IP_PREFIX"
+    IP_RANGE_FROM_BINARY = "IP_RANGE_FROM_BINARY"
+    IP_RANGE_TO_BINARY = "IP_RANGE_TO_BINARY"
 
 
 class SearchKey(Enum):
@@ -60,6 +64,10 @@ class SearchKey(Enum):
     IPV6_ADDRESS = FileColumnMeaningType.IPV6_ADDRESS
     IPV6_RANGE_FROM = FileColumnMeaningType.IPV6_RANGE_FROM
     IPV6_RANGE_TO = FileColumnMeaningType.IPV6_RANGE_TO
+    IP_BINARY = FileColumnMeaningType.IP_BINARY
+    IP_RANGE_FROM_BINARY = FileColumnMeaningType.IP_RANGE_FROM_BINARY
+    IP_RANGE_TO_BINARY = FileColumnMeaningType.IP_RANGE_TO_BINARY
+    IP_PREFIX = FileColumnMeaningType.IP_PREFIX
 
     # For data source registration. Don't use it for FeaturesEnricher
     EMAIL_ONE_DOMAIN = FileColumnMeaningType.EMAIL_ONE_DOMAIN
@@ -81,7 +89,7 @@ class SearchKey(Enum):
         if meaning_type == FileColumnMeaningType.EMAIL:
             return SearchKey.EMAIL
         if meaning_type == FileColumnMeaningType.HEM:
-            return SearchKey.HEM
+            return SearchKey.HEM  # TODO check that it wasn't EMAIL
         if meaning_type == FileColumnMeaningType.IP_ADDRESS:
             return SearchKey.IP
         if meaning_type == FileColumnMeaningType.MSISDN:
@@ -97,21 +105,42 @@ class SearchKey(Enum):
         if meaning_type == FileColumnMeaningType.POSTAL_CODE:
             return SearchKey.POSTAL_CODE
         if meaning_type == FileColumnMeaningType.IPV6_ADDRESS:
-            return SearchKey.IPV6_ADDRESS
-        if meaning_type == FileColumnMeaningType.IPV6_RANGE_FROM:
-            return SearchKey.IPV6_RANGE_FROM
-        if meaning_type == FileColumnMeaningType.IPV6_RANGE_TO:
-            return SearchKey.IPV6_RANGE_TO
-        if meaning_type == FileColumnMeaningType.EMAIL_ONE_DOMAIN:
-            return SearchKey.EMAIL_ONE_DOMAIN
-        if meaning_type == FileColumnMeaningType.IP_RANGE_FROM:
-            return SearchKey.IP_RANGE_FROM
-        if meaning_type == FileColumnMeaningType.IP_RANGE_TO:
-            return SearchKey.IP_RANGE_TO
-        if meaning_type == FileColumnMeaningType.MSISDN_RANGE_FROM:
-            return SearchKey.MSISDN_RANGE_FROM
-        if meaning_type == FileColumnMeaningType.MSISDN_RANGE_TO:
-            return SearchKey.MSISDN_RANGE_TO
+            return SearchKey.IP
+        # if meaning_type == FileColumnMeaningType.IPV6_RANGE_FROM:
+        #     return SearchKey.IPV6_RANGE_FROM
+        # if meaning_type == FileColumnMeaningType.IPV6_RANGE_TO:
+        #     return SearchKey.IPV6_RANGE_TO
+        # if meaning_type == FileColumnMeaningType.EMAIL_ONE_DOMAIN:
+        #     return SearchKey.EMAIL_ONE_DOMAIN
+        # if meaning_type == FileColumnMeaningType.IP_RANGE_FROM:
+        #     return SearchKey.IP_RANGE_FROM
+        # if meaning_type == FileColumnMeaningType.IP_RANGE_TO:
+        #     return SearchKey.IP_RANGE_TO
+        # if meaning_type == FileColumnMeaningType.MSISDN_RANGE_FROM:
+        #     return SearchKey.MSISDN_RANGE_FROM
+        # if meaning_type == FileColumnMeaningType.MSISDN_RANGE_TO:
+        #     return SearchKey.MSISDN_RANGE_TO
+        if meaning_type == FileColumnMeaningType.IP_BINARY:
+            return SearchKey.IP
+        # if meaning_type == FileColumnMeaningType.IP_RANGE_FROM_BINARY:
+        #     return SearchKey.IP_RANGE_FROM_BINARY
+        # if meaning_type == FileColumnMeaningType.IP_RANGE_TO_BINARY:
+        #     return SearchKey.IP_RANGE_TO_BINARY
+
+    @staticmethod
+    def find_key(search_keys: Dict[str, SearchKey], keys: Union[SearchKey, List[SearchKey]]) -> Optional[SearchKey]:
+        if isinstance(keys, SearchKey):
+            keys = [keys]
+        for col, key_type in search_keys.items():
+            if key_type in keys:
+                return col
+        return None
+
+    @staticmethod
+    def find_all_keys(search_keys: Dict[str, SearchKey], keys: Union[SearchKey, List[SearchKey]]) -> List[SearchKey]:
+        if isinstance(keys, SearchKey):
+            keys = [keys]
+        return [col for col, key_type in search_keys.items() if key_type in keys]
 
 
 class DataType(Enum):
@@ -121,6 +150,7 @@ class DataType(Enum):
     DATE_TIME = "DATE_TIME"
     STRING = "STRING"
     BOOLEAN = "BOOLEAN"
+    BYTES = "BYTES"
 
 
 class ModelTaskType(Enum):
@@ -129,6 +159,18 @@ class ModelTaskType(Enum):
     REGRESSION = "REGRESSION"
     TIMESERIES = "TIMESERIES"
 
+    def is_classification(self) -> bool:
+        return self in [ModelTaskType.BINARY, ModelTaskType.MULTICLASS]
+
+    @staticmethod
+    def parse(task_type: Any) -> "ModelTaskType":
+        if isinstance(task_type, ModelTaskType):
+            return task_type
+        elif isinstance(task_type, str):
+            return ModelTaskType(task_type.upper())
+        else:
+            raise ValueError(f"Invalid task type: {task_type}")
+
 
 class ModelLabelType(Enum):
     GINI = "gini"
@@ -157,23 +199,23 @@ class FileMetricsInterval(BaseModel):
     date_cut: float
     count: float
     valid_count: float
-    avg_target: Optional[float]  # not for multiclass
-    avg_score_etalon: Optional[float]
+    avg_target: Optional[float] = None  # not for multiclass
+    avg_score_etalon: Optional[float] = None
 
 
 class FileMetrics(BaseModel):
     # etalon metadata
-    task_type: Optional[ModelTaskType]
-    label: Optional[ModelLabelType]
-    count: Optional[int]
-    valid_count: Optional[int]
-    valid_rate: Optional[float]
-    avg_target: Optional[float]
-    metrics_binary_etalon: Optional[BinaryTask]
-    metrics_regression_etalon: Optional[RegressionTask]
-    metrics_multiclass_etalon: Optional[MulticlassTask]
-    cuts: Optional[List[float]]
-    interval: Optional[List[FileMetricsInterval]]
+    task_type: Optional[ModelTaskType] = None
+    label: Optional[ModelLabelType] = None
+    count: Optional[int] = None
+    valid_count: Optional[int] = None
+    valid_rate: Optional[float] = None
+    avg_target: Optional[float] = None
+    metrics_binary_etalon: Optional[BinaryTask] = None
+    metrics_regression_etalon: Optional[RegressionTask] = None
+    metrics_multiclass_etalon: Optional[MulticlassTask] = None
+    cuts: Optional[List[float]] = None
+    interval: Optional[List[FileMetricsInterval]] = None
 
 
 class NumericInterval(BaseModel):
@@ -187,25 +229,25 @@ class FileColumnMetadata(BaseModel):
     dataType: DataType
     meaningType: FileColumnMeaningType
     minMaxValues: Optional[NumericInterval] = None
-    originalName: Optional[str]
+    originalName: Optional[str] = None
     # is this column contains keys from multiple key columns like msisdn1, msisdn2
     isUnnest: bool = False
     # list of original etalon key column names like msisdn1, msisdn2
-    unnestKeyNames: Optional[List[str]]
+    unnestKeyNames: Optional[List[str]] = None
 
 
 class FileMetadata(BaseModel):
     name: str
-    description: Optional[str]
+    description: Optional[str] = None
     columns: List[FileColumnMetadata]
     searchKeys: List[List[str]]
-    excludeFeaturesSources: Optional[List[str]]
-    hierarchicalGroupKeys: Optional[List[str]]
-    hierarchicalSubgroupKeys: Optional[List[str]]
-    taskType: Optional[ModelTaskType]
-    rowsCount: Optional[int]
-    checksumMD5: Optional[str]
-    digest: Optional[str]
+    excludeFeaturesSources: Optional[List[str]] = None
+    hierarchicalGroupKeys: Optional[List[str]] = None
+    hierarchicalSubgroupKeys: Optional[List[str]] = None
+    taskType: Optional[ModelTaskType] = None
+    rowsCount: Optional[int] = None
+    checksumMD5: Optional[str] = None
+    digest: Optional[str] = None
 
     def column_by_name(self, name: str) -> Optional[FileColumnMetadata]:
         for c in self.columns:
@@ -213,13 +255,15 @@ class FileMetadata(BaseModel):
                 return c
         return None
 
-    def search_types(self) -> Set[SearchKey]:
-        search_keys = set()
+    def search_types(self) -> Dict[SearchKey, str]:
+        search_keys = dict()
         for keys_group in self.searchKeys:
             for key in keys_group:
                 column = self.column_by_name(key)
                 if column:
-                    search_keys.add(SearchKey.from_meaning_type(column.meaningType))
+                    search_key = SearchKey.from_meaning_type(column.meaningType)
+                    if search_key is not None:
+                        search_keys[search_key] = column.name
         return search_keys
 
 
@@ -229,17 +273,18 @@ class FeaturesMetadataV2(BaseModel):
     source: str
     hit_rate: float
     shap_value: float
-    commercial_schema: Optional[str]
-    data_provider: Optional[str]
-    data_providers: Optional[List[str]]
-    data_provider_link: Optional[str]
-    data_provider_links: Optional[List[str]]
-    data_source: Optional[str]
-    data_sources: Optional[List[str]]
-    data_source_link: Optional[str]
-    data_source_links: Optional[List[str]]
-    doc_link: Optional[str]
-    update_frequency: Optional[str]
+    commercial_schema: Optional[str] = None
+    data_provider: Optional[str] = None
+    data_providers: Optional[List[str]] = None
+    data_provider_link: Optional[str] = None
+    data_provider_links: Optional[List[str]] = None
+    data_source: Optional[str] = None
+    data_sources: Optional[List[str]] = None
+    data_source_link: Optional[str] = None
+    data_source_links: Optional[List[str]] = None
+    doc_link: Optional[str] = None
+    update_frequency: Optional[str] = None
+    from_online_api: Optional[bool] = None
 
 
 class HitRateMetrics(BaseModel):
@@ -259,48 +304,52 @@ class ModelEvalSet(BaseModel):
 class BaseColumnMetadata(BaseModel):
     original_name: str
     hashed_name: str
-    ads_definition_id: Optional[str]
+    ads_definition_id: Optional[str] = None
     is_augmented: bool
 
 
 class GeneratedFeatureMetadata(BaseModel):
-    alias: Optional[str]
+    alias: Optional[str] = None
     formula: str
     display_index: str
     base_columns: List[BaseColumnMetadata]
-    operator_params: Optional[Dict[str, str]]
+    operator_params: Optional[Dict[str, str]] = None
 
 
 class ProviderTaskMetadataV2(BaseModel):
     features: List[FeaturesMetadataV2]
-    hit_rate_metrics: Optional[HitRateMetrics]
-    eval_set_metrics: Optional[List[ModelEvalSet]]
-    zero_hit_rate_search_keys: Optional[List[str]]
-    features_used_for_embeddings: Optional[List[str]]
-    shuffle_kfold: Optional[bool]
-    generated_features: Optional[List[GeneratedFeatureMetadata]]
+    hit_rate_metrics: Optional[HitRateMetrics] = None
+    eval_set_metrics: Optional[List[ModelEvalSet]] = None
+    zero_hit_rate_search_keys: Optional[List[str]] = None
+    features_used_for_embeddings: Optional[List[str]] = None
+    shuffle_kfold: Optional[bool] = None
+    generated_features: Optional[List[GeneratedFeatureMetadata]] = None
 
 
 class FeaturesFilter(BaseModel):
-    minImportance: Optional[float]
-    maxPSI: Optional[float]
-    maxCount: Optional[int]
-    selectedFeatures: Optional[List[str]]
+    minImportance: Optional[float] = None
+    maxPSI: Optional[float] = None
+    maxCount: Optional[int] = None
+    selectedFeatures: Optional[List[str]] = None
 
 
 class RuntimeParameters(BaseModel):
-    properties: Dict[str, str] = {}
+    properties: Dict[str, Any] = {}
+
+
+class AutoFEParameters(BaseModel):
+    ts_gap_days: int = 0
 
 
 class SearchCustomization(BaseModel):
-    featuresFilter: Optional[FeaturesFilter]
-    extractFeatures: Optional[bool]
-    accurateModel: Optional[bool]
-    importanceThreshold: Optional[float]
-    maxFeatures: Optional[int]
-    returnScores: Optional[bool]
-    runtimeParameters: Optional[RuntimeParameters]
-    metricsCalculation: Optional[bool]
+    featuresFilter: Optional[FeaturesFilter] = None
+    extractFeatures: Optional[bool] = None
+    accurateModel: Optional[bool] = None
+    importanceThreshold: Optional[float] = None
+    maxFeatures: Optional[int] = None
+    returnScores: Optional[bool] = None
+    runtimeParameters: Optional[RuntimeParameters] = None
+    metricsCalculation: Optional[bool] = None
 
     def __repr__(self):
         return (
@@ -321,3 +370,6 @@ class CVType(Enum):
     time_series = "time_series"
     blocked_time_series = "blocked_time_series"
     not_set = "not_set"
+
+    def is_time_series(self) -> bool:
+        return self in [CVType.time_series, CVType.blocked_time_series]
